



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/videodev2.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/mman.h>
#include "mprintf.h"



#include "capture.h"

 static  void choose_format(vcdev_t *D)
; static  void unmap_buffers(vcdev_t *D)
; static  char *fourcc2str(int X)
; static  unsigned int yuv2rgb(int y, int u, int v)

; static  void cvt_YUYV_RGB32(vcdev_t *D,int fno, unsigned *outp) 
;


void *
__unused_capture_c

	[]= {
	
	0,
};












 vcdev_t* open_device(char *devname)
{
	int i;
	vcdev_t *D;
	struct v4l2_capability caps;

	
	D= calloc(1, sizeof(D[0]));


	D->devname= strdup(devname);
	D->fd= open(devname, O_RDWR);
	if (D->fd<0) {
		D->errcode= errno;
		D->error_msg= mprintf("can not open device %s: %s",
			devname, strerror(errno));
		return D;
	}
	
	if (ioctl(D->fd, VIDIOC_QUERYCAP, &caps))
	{
		/* FIXME: handle EINVAL */
		D->errcode= errno;
		D->error_msg= mprintf("can not get capabilities of device %s:"
		 " %s", devname, strerror(D->errcode));
		return D;
	}

	if ((caps.capabilities & V4L2_CAP_VIDEO_CAPTURE)==0) {
		D->errcode= EOPNOTSUPP;
		D->error_msg= mprintf("device %s does not "
		   "support video capture",devname);
		return D;
	}

	D->method= 0;
	if (caps.capabilities & V4L2_CAP_STREAMING)
		D->method= 'M';
	else if (caps.capabilities & V4L2_CAP_READWRITE) 
		D->method= 'R';

	if (D->method==0) {
		D->errcode= EOPNOTSUPP;
		D->error_msg= mprintf("device %s does not support "
		   "neither mmap nor read I/O methods. ", devname);
		return D;
	}

	/* OMITTING SECTIONS 1.4 1.5 1.6 1.7 1.8 1.9 */

	choose_format(D);
	if (D->errcode) return D;


	/* OMITTING THE COLORSPACE STUFF */
	/* OMITTING SECTIONS 1.11 1.12 */


	if (D->method=='M') {
		struct v4l2_requestbuffers reqb;
		memset(&reqb, 0, sizeof(reqb));
		reqb.count= 4; // FIXME: make it configurable
		reqb.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
		reqb.memory= V4L2_MEMORY_MMAP;

		if (ioctl(D->fd, VIDIOC_REQBUFS, &reqb))
		{
			if (errno==EINVAL) goto try_user_ptr_method;
			D->errcode= errno;
			D->error_msg= mprintf("can not allocate buffers "
				"on device %s: %s", D->devname, 
					strerror(errno));
			return D;
		}

		if (reqb.count==0) {
			D->errcode= ENOMEM;
			D->error_msg= mprintf("device %s is out of memory. "
				"can not allocate buffers on it.",D->devname);
			return D;
		}

		D->nbuffers= reqb.count;
		D->buffers= malloc(sizeof(D->buffers[0])*D->nbuffers);
		for(i=0;i<D->nbuffers;i++)
			D->buffers[i].mapped= 0;

		for(i=0;i<D->nbuffers;i++) {
			struct v4l2_buffer buf;
			memset(&buf,0,sizeof(buf));
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_MMAP;
			buf.index = i;

			if (ioctl(D->fd, VIDIOC_QUERYBUF, &buf)) {
				D->errcode= errno;
				D->error_msg= mprintf("can not access "
				  "allocated buffer %d on device %s: %s",
				i, D->devname, strerror(errno));
				goto cleanup_mmaps;
			}
			D->buffers[i].length= buf.length;
			D->buffers[i].address= mmap(NULL, buf.length,
				PROT_READ | PROT_WRITE , MAP_SHARED,
				D->fd, buf.m.offset);
			if (!D->buffers[i].address) {
				D->errcode= errno;
				D->error_msg= mprintf("can not mmap buffer %d "
					"of length %d on device %s: %s",
					i, buf.length, D->devname, 
					strerror(errno));
				goto cleanup_mmaps;
			}
			D->buffers[i].length= buf.length;
			D->buffers[i].mapped= 1;
		}
	}

	

	/* I think all is done, eh? */
	return D;

try_user_ptr_method:
	D->errcode= EOPNOTSUPP;
	D->error_msg= mprintf("device %s does not support mmap()ed I/O "
		"but the program can not handle user pointer stuff.");
	return D;

cleanup_mmaps:
	unmap_buffers(D);
	return D;
}


static int usable_formats[]= {
	V4L2_PIX_FMT_YUYV,
	-1,
};

 static  void choose_format(vcdev_t *D)
{
	int i;
	struct v4l2_fmtdesc desc;
	struct v4l2_format fmt;
	int j;
	int best_at;

	best_at= -1;
	for(i=0;;i++) {
		memset(&desc,0,sizeof(desc));
		desc.index= i;
		desc.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (ioctl(D->fd, VIDIOC_ENUM_FMT, &desc)) 
			/* ignore errors, since all we do is query, yes? */
			break;
		for(j=0;usable_formats[j]!=-1;j++)
			if (usable_formats[j]==desc.pixelformat) {
				if (best_at==-1 || j<best_at) best_at= j;
				break;
			}
	}

	if (best_at==-1)
	{
		D->errcode= EINVAL;
		D->error_msg= mprintf("no suitable formats were found "
			"on device %s", D->devname);
		return ;
	}

	memset(&fmt,0,sizeof(fmt));
	fmt.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.pixelformat= usable_formats[best_at];
	// FIXME: these should be adjustable
	fmt.fmt.pix.width= 1000;
	fmt.fmt.pix.height= 1000;
	if (ioctl(D->fd,VIDIOC_S_FMT, &fmt)) {
		D->errcode= errno;
		D->error_msg= mprintf("can not set format %s on "
			"device %s: %s", fourcc2str(usable_formats[best_at]),
				D->devname, strerror(errno));
		return ;
	}
	
	D->width= fmt.fmt.pix.width;
	D->height= fmt.fmt.pix.height;
	D->pixelformat= fmt.fmt.pix.pixelformat;
	D->bytesperline= fmt.fmt.pix.bytesperline;
	D->sizeimage= fmt.fmt.pix.sizeimage;

	if (D->pixelformat!=usable_formats[best_at]) {
		for(i=0;usable_formats[i]!=-1;i++)
			if (usable_formats[i]==D->pixelformat)
				break;
		if (usable_formats[i]==-1) {
			D->errcode= EINVAL;
			D->error_msg= mprintf("requested format %s, "
				"but device %s returned unsupported "
				"format %s.",
					fourcc2str(usable_formats[best_at]),
					D->devname,
					fourcc2str(D->pixelformat));
			return ;
		}
	}
}



 static  void unmap_buffers(vcdev_t *D)
{
	int i;
	for(i=0;i<D->nbuffers;i++) {
		if (D->buffers[i].mapped) 
			munmap(D->buffers[i].address, D->buffers[i].length);
	}
	free(D->buffers);
	D->nbuffers= 0;
}

 void start_capture(vcdev_t *D)
{
	int i;
        enum v4l2_buf_type type;

	if (D->method!='M') return ;

	for(i=0;i<D->nbuffers;i++) {
		putback_buffer(D,i);
		if (D->errcode) return ;
	}

	type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(D->fd, VIDIOC_STREAMON, &type)) {
		D->errcode= errno;
		D->error_msg= mprintf("can not start video stream on device "
			"%s: %s", D->devname, strerror(errno));
		return ;
	}
	D->errcode= 0;
}

 int read_frame(vcdev_t *D)
{
	struct v4l2_buffer buf;	

	buf.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory= V4L2_MEMORY_MMAP;
	if (ioctl(D->fd, VIDIOC_DQBUF, &buf)) {
		D->errcode= errno;
		D->error_msg= mprintf("can not read frame from device %s: "
				"%s", D->devname, strerror(errno));
		return -1;
	}
	if (buf.index<0||buf.index>=D->nbuffers) {
		D->errcode= EINVAL;
		D->error_msg= mprintf("invalid buffer id %d was returned "
			"by device %s: should be between 0 and %d",
				buf.index, D->devname, D->nbuffers-1);
		return -1;
	}
	D->errcode= 0;
	return buf.index;
} 

 void putback_buffer(vcdev_t *D,int index)
{
	struct v4l2_buffer buf;
	memset(&buf,0, sizeof(buf));

	buf.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory= V4L2_MEMORY_MMAP;
	buf.index= index;

	if (ioctl (D->fd, VIDIOC_QBUF, &buf)) {
		D->errcode= errno;
		D->error_msg= mprintf("can not queue buffer %d on "
			"device %s: %s", index, D->devname,
			strerror(errno));
		return ;
	}
}

 void stop_capture(vcdev_t *D)
{
        enum v4l2_buf_type type;
	type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(D->fd, VIDIOC_STREAMOFF, &type)) {
		D->errcode= errno;
		D->error_msg= mprintf("can not stop video capture on "
			"device %s: %s", D->devname, strerror(errno));
	}
}

 void close_device(vcdev_t *D)
{
	unmap_buffers(D);
	if (D->fd>=0) close(D->fd);
	free(D->devname);
	if (D->error_msg) free(D->error_msg);
	free(D);
}

 void enumerate_formats(vcdev_t *D)
{
	int i;
	struct v4l2_fmtdesc desc;

	if (D->fd<0) {
		D->errcode= EINVAL;
		D->error_msg= mprintf("device %s was not initialized "
			"properly, can not query formats.");
		return ;
	}
	for(i=0;;i++) {
		memset(&desc,0,sizeof(desc));
		desc.index= i;
		desc.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (ioctl(D->fd, VIDIOC_ENUM_FMT, &desc)) {
			if (errno==EINVAL) {
				if (i==0)
					printf("No formats reported "
					"by device %s\n", D->devname);
				break;
			} else {
				D->errcode= errno;
				D->error_msg= mprintf("error during query "
				   "of supported formats on device %s: %s",
					D->devname, strerror(errno)); 
				break;
			}
		}
		if (i==0)
			printf("Image formats supported by device %s:\n",
				D->devname);
		printf("%s%s\n", fourcc2str(desc.pixelformat),
			(desc.flags & V4L2_FMT_FLAG_COMPRESSED) ?
			" compressed" : "");
	}
}

 static  char *fourcc2str(int X)
{
	char b[5];
	*(int*) b= X;
	b[4]= 0;
	return strdup(b);
}

 static  unsigned int yuv2rgb(int y, int u, int v)

{
   unsigned int pixel32;
   unsigned char *pixel = (unsigned char *)&pixel32;
   int r, g, b;

   r = y + (1.370705 * (v-128));
   g = y - (0.698001 * (v-128)) - (0.337633 * (u-128));
   b = y + (1.732446 * (u-128));

   if (r > 255) r = 255;
   if (g > 255) g = 255;
   if (b > 255) b = 255;

   if (r < 0) r = 0;
   if (g < 0) g = 0;
   if (b < 0) b = 0;

   // Values only go from 0-220..  Why?

   pixel[0] = b * 220 / 256;
   pixel[1] = g * 220 / 256;
   pixel[2] = r * 220 / 256;
   pixel[3] = 0;

   return pixel32;
}





 static  void cvt_YUYV_RGB32(vcdev_t *D,int fno, unsigned *outp) 
{
	int i;
	unsigned char *inp;
	int y;

	inp= (unsigned char*) D->buffers[fno].address;

#if 0

	printf("buffer length= %d, required= %d\n",D->buffers[fno].length,
		D->width*2*D->height);
	
	printf("Skip: %d\n",D->bytesperline - D->width*2);
#endif

	for(y=0;y<D->height;y++) {
		int Y1,Y2, Cb, Cr;
		for(i=0;i<D->width/2;i++) {
			Y1= *inp; inp++;
			Cb= *inp; inp++;	
			Y2= *inp; inp++;
			Cr= *inp; inp++;

			*outp= yuv2rgb(Y1,Cb,Cr); outp++;
			*outp= yuv2rgb(Y2,Cb,Cr); outp++;
		}
		inp+= D->bytesperline - D->width*2;
	}
}

 unsigned char *grey_buffer(vcdev_t *D)
{
	unsigned char *Z;
	Z= malloc(D->width*D->height);
	return Z;
}

 unsigned int* rgb24_buffer(vcdev_t *D)
{
	unsigned int *Z;
	Z= malloc(D->width*D->height*sizeof(unsigned));
	return Z;
}

 void cvt_rgb24(vcdev_t *D,int fno, unsigned int *out)
{
	cvt_YUYV_RGB32(D, fno,out);	
}

 void cvt_grey(vcdev_t *D,int fno, unsigned char *out)
{
	int y,i;
	unsigned char *inp;

	inp= (unsigned char*) D->buffers[fno].address;

	for(y=0;y<D->height;y++) {
		int Y1,Y2, Cb, Cr;
		for(i=0;i<D->width/2;i++) {
			*out= *inp; inp++; out++;
			/* skip Cb */ inp++;	
			*out= *inp; inp++; out++;
			/* skip Cr */ inp++; 
		}
		inp+= D->bytesperline - D->width*2;
	}
}


